import itertools
import os
import logging
import time

import requests

from ruv_dl.data import Entry, EntrySet
from ruv_dl.programs import ProgramInfo
from ruv_dl.constants import PROGRAM_INFO_FN
from ruv_dl.migrations import MIGRATIONS

logger = logging.getLogger(__name__)
PROGRAM_INFO_VERSION = max(MIGRATIONS.keys())


class Downloader:
    def __init__(self, destination, program, episode_entries, threaded=True):
        self.destination = destination
        self.program = program
        self.episode_entries = episode_entries
        self.threaded = threaded

    def organize(self):
        # TODO: Use ProgramInfo class
        logger.info(f'Organizing {self.program["title"]}')
        info_fn = os.path.join(
            self.destination,
            self.program['title'],
            PROGRAM_INFO_FN,
        )
        os.makedirs(os.path.dirname(info_fn), exist_ok=True)
        try:
            program_info = ProgramInfo(info_fn)
        except FileNotFoundError:
            program_info = ProgramInfo(info_fn, initialize_empty=True)
        seasons = program_info.seasons
        program_info.program = self.program
        # seasons = {
        #     1: {entry, entry, entry},
        #     2: {entry, entry, entry},
        # }
        # Sort episodes into seasons
        for entry in sorted(
            self.episode_entries,
            key=lambda entry: entry.date
        ):
            for season in seasons.keys():
                if any(
                    abs((e.date - entry.date).days) < 10
                    for e in seasons[season]
                ):
                    seasons[season].add(entry)
                    break
            else:
                season = max((seasons or {0: 0}).keys()) + 1
                seasons[season] = EntrySet([entry])
        # Calculate target paths for entries
        for season, entries in seasons.items():
            season_folder = Entry.get_season_folder(
                self.destination, self.program, season
            )
            os.makedirs(season_folder, exist_ok=True)
            for i, entry in enumerate(entries.sorted()):
                if not entry.episode.number:
                    entry.episode.number = EntrySet.find_target_number(
                        entries, i
                    )
                basename = entry.get_target_basename(
                    self.program,
                    season,
                )
                target_path = os.path.join(
                    season_folder,
                    basename,
                )
                entry.set_target_path(target_path)
        # Finally, make sure we don't have the same etag multiple times,
        # prefer the first one in chronological order
        found_etags = []
        for season, entries in seasons.items():
            for entry in [
                entry for entry in entries if entry.etag in found_etags
            ]:
                entries.remove(entry)
            found_etags += [entry.etag for entry in entries]

        program_info.seasons = seasons
        program_info.write()

        missing_migrations = range(
            program_info.version,
            PROGRAM_INFO_VERSION,
        )
        for migration_entry in missing_migrations:
            logger.error(
                'Missing migration %d. Run `ruv-dl migrate %d`. You can '
                'supply `--dryrun` (e.g. `ruv-dl --dryrun migrate ...`) to '
                'see what will be done.',
                migration_entry + 1,
                migration_entry + 1,
            )
        if missing_migrations:
            return []

        return [
            entry
            for entry in itertools.chain(*seasons.values())
            if not entry.exists_on_disk()
        ]

    def download_file(self, entry):
        if os.path.exists(entry.target_path):
            logger.info(
                f'Skipping {entry.target_path} - {entry.url} because '
                'it already exists.'
            )
            return False
        else:
            logger.warning(f'Downloading {entry.url} to {entry.target_path}')

        r = requests.get(entry.url, stream=True)

        if r.ok:
            start = time.time()
            total_length = int(r.headers.get('content-length'))
            dl = 0
            perc_done = 0
            with open(entry.target_path, 'wb') as f:
                for chunk in r:
                    dl += len(chunk)
                    current = int(dl * 10 / total_length)
                    if current > perc_done:
                        perc_done = current
                        logger.info(
                            f'{os.path.basename(entry.target_path)} '
                            f'{perc_done * 10}% '
                            f'({int(dl//(time.time() - start)/1024)}kbps)'
                        )
                    f.write(chunk)

            size = int(os.path.getsize(entry.target_path) / 1024**2)
            logger.warning(
                f'{entry.target_path} ({size}MB) '
                f'downloaded in {int(time.time() - start)}s!'
            )
            return True
        logger.warning(f'Error {r.status_code} for {entry.url}')
        return False



#!/usr/bin/env python
import itertools
import os
import shutil
import logging
import sys
import click
import multiprocessing
from multiprocessing.pool import ThreadPool

from ruv_dl.programs import ProgramFetcher
from ruv_dl.crawler import Crawler
from ruv_dl.downloader import Downloader
from ruv_dl.migrations import MIGRATIONS
from ruv_dl.constants import DEFAULT_VIDEO_DESTINATION, CACHE_LOCATION


logger = logging.getLogger('ruv_dl')
handler = logging.StreamHandler(sys.stdout)
format_str = '%(asctime)s - %(name)s -  %(message)s'
date_format = '%Y-%m-%d %H:%M:%S'
formatter = logging.Formatter(format_str, date_format)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.WARN)


@click.group()
@click.option('--dryrun/--no-dryrun', default=False)
@click.option('-v', '--verbosity', count=True)
@click.option('--empty-cache', default=False, is_flag=True)
@click.option(
    '-d', '--destination', default=DEFAULT_VIDEO_DESTINATION,
    type=click.Path(),
    help='Top level destination directory.',
)
@click.pass_context
def cli(ctx, dryrun, verbosity, empty_cache, destination):
    ctx.obj['dryrun'] = dryrun
    ctx.obj['destination'] = destination
    if verbosity is not None:
        multiprocessing_logger = multiprocessing.get_logger()
        if verbosity > 2:
            multiprocessing_logger.addHandler(handler)
            multiprocessing_logger.setLevel(logging.DEBUG)
        if verbosity > 1:
            logger.setLevel(logging.DEBUG)
            multiprocessing_logger.addHandler(handler)
            multiprocessing_logger.setLevel(logging.INFO)
        elif verbosity > 0:
            logger.setLevel(logging.INFO)

    if empty_cache:
        if os.path.exists(CACHE_LOCATION):
            shutil.rmtree(CACHE_LOCATION)
    os.makedirs(CACHE_LOCATION, exist_ok=True)


@cli.command()
@click.argument(
    'query',
    nargs=-1,
    type=click.STRING,
)
@click.option(
    '-u', '--update', default=False, is_flag=True,
    help='Search for all saved shows in `--destination` and download '
    'available episodes',
)
@click.option(
    '--days-between-episodes', type=click.INT, default=7,
    help='Rate of episode release',
)
@click.option(
    '--iteration-count', type=click.INT, default=5,
    help='Maximum passes to allow for no shows found.',
)
@click.option(
    '--sequential', default=False, is_flag=True,
    help='Do not run threaded, only download one file at a time.',
)
@click.pass_context
def download(
    ctx, query, update, days_between_episodes, iteration_count,
    sequential,
):
    '''
        Download ruv programs by searching for query (can specify multiple)
        or update your currently synced programs.
    '''
    destination = ctx.obj['destination']
    if bool(query) == bool(update):
        click.echo(download.get_help(ctx))
        raise click.UsageError(
            'Query terms and update are mutually exclusive and either must '
            'be included'
        )
    os.makedirs(destination, exist_ok=True)
    fetcher = ProgramFetcher(query, update, destination)
    with ThreadPool(8) as pool:
        programs = {}
        for program in fetcher.get_programs():
            logger.info(f'------ {program["title"]} [{program["id"]}] ------')
            crawler = Crawler(
                days_between_episodes=days_between_episodes,
                iteration_count=iteration_count,
                program=program,
            )
            programs[program['id']] = {
                'program': program,
                'episodes': pool.apply_async(
                    crawler.search_for_episodes
                )
            }

        downloaders = []
        for program_id, data in programs.items():
            downloader = Downloader(
                destination=destination,
                program=data['program'],
                episode_entries=data['episodes'].get(),
                threaded=not sequential,
            )
            entries = downloader.organize()
            downloaders.append((downloader, entries))

    total_entries_to_download = sum(
        len(entries) for _, entries in downloaders
    )
    if not total_entries_to_download:
        logger.info('No entries to download, bye')
    else:
        logger.warning(f'Downloading {total_entries_to_download} files...')
        if ctx.obj['dryrun']:
            logger.warning('Dryrun, not downloading anything, bye')
            return
        if sequential:
            results = [
                [
                    downloader.download_file(entry)
                    for entry in entries
                ] for downloader, entries in downloaders
            ]
        else:
            with ThreadPool(8) as pool:
                pool.__exit__
                async_results = [
                    pool.map_async(downloader.download_file, entries)
                    for downloader, entries in downloaders
                ]
                results = [result.get() for result in async_results]
        logger.warning(
            f'{len([r for r in itertools.chain(*results) if r])} '
            'files downloaded'
        )


@cli.command()
@click.argument('migration', type=click.INT)
@click.pass_context
def migrate(ctx, migration):
    for entry in MIGRATIONS[migration]:
        entry(dryrun=ctx.obj['dryrun'], destination=ctx.obj['destination'])


def main():
    cli(obj={})



#!/usr/bin/env python
import json
import os
import logging

from ruv_dl.constants import CACHE_LOCATION, CACHE_VERSION, CACHE_VERSION_KEY

logger = logging.getLogger(__name__)


class CacheVersionException(Exception):
    pass


class DiskCache:
    def __init__(self, program_id):
        self.location = os.path.join(CACHE_LOCATION, f'{program_id}.json')
        try:
            with open(self.location, 'r') as f:
                self._data = json.loads(f.read())
            SAVED_CACHE_VERSION = self._data.get(CACHE_VERSION_KEY)
            if SAVED_CACHE_VERSION != CACHE_VERSION:
                logger.info(
                    f'Have cache version "{SAVED_CACHE_VERSION}" but '
                    f'want {CACHE_VERSION}. Starting with empty cache.'
                )
                raise CacheVersionException()
            logger.debug('Cache version OK.')
        except (FileNotFoundError, CacheVersionException):
            self._data = {
                CACHE_VERSION_KEY: CACHE_VERSION,
            }

    def get(self, key):
        return self._data[key]

    def set(self, key, data):
        self._data[key] = data

    def has(self, key):
        return key in self._data

    def remove(self, key):
        del self._data[key]

    def write(self):
        with open(self.location, 'w') as f:
            f.write(json.dumps(self._data))



import logging
import os
import copy
import shutil

from ruv_dl.programs import ProgramFetcher
from ruv_dl.data import Entry, Episode

logger = logging.getLogger(__name__)

'''
    TODO: Split migrations to seperate files when/if we need more.
          Create some kind of autodiscovery and autopopulation of migrations.
          Also automatically bump version in program_info?
'''


def move_old_locations(destination, dryrun=False):
    fetcher = ProgramFetcher(None, None, destination)
    for program_info in fetcher.get_all_program_infos():
        if program_info.version >= 1:
            logger.info('Skipping %s', program_info)
            continue
        program = program_info.program
        logger.info('Targeting %s', program['title'])
        seasons = program_info.seasons
        # {current_path: new_path}
        files_to_move = {}
        for season, entries in seasons.items():
            season_folder = Entry.get_season_folder(
                destination, program, season
            )
            for i, target_entry in enumerate(entries.sorted()):
                if target_entry.episode.number is None:
                    raise RuntimeError(
                        'You need to attempt sync for this program once '
                        'before running this migration.'
                    )
                src_entry = copy.deepcopy(target_entry)
                src_entry.episode = Episode(None)
                src_entry.episode.number = i + 1

                src_dest = os.path.join(
                    season_folder,
                    src_entry.get_target_basename(program, season),
                )

                target_dest = os.path.join(
                    season_folder,
                    target_entry.get_target_basename(program, season),
                )

                if os.path.isfile(src_dest):
                    files_to_move[src_dest] = target_dest
        if dryrun:
            logger.warning('Running with dryrun, no change to filesystem')
            for current, target in files_to_move.items():
                logger.info('Would move %s to %s', current, target)
        else:
            for current, target in files_to_move.items():
                os.makedirs(os.path.dirname(target), exist_ok=True)
                logger.info('Moving %s to %s', current, target)
                shutil.move(current, target)
            program_info.version = 1
            program_info.write()


MIGRATIONS = {
    1: (move_old_locations, )
}



#!/usr/bin/env python
import json
import os
import logging
import glob

import requests

from ruv_dl.data import Entry, EntrySet
from ruv_dl.constants import PROGRAM_INFO_FN, NON_SEASON_FIELDS

logger = logging.getLogger(__name__)


class ProgramInfo:
    def __init__(self, fn, initialize_empty=False):
        self.fn = fn
        if initialize_empty:
            self._data = {
                '__version__': 1
            }
        else:
            with open(fn, 'r') as f:
                try:
                    data = json.loads(f.read())
                except ValueError:
                    logger.info('Could not parse %s', fn)
                else:
                    if 'program' in data:
                        if 'id' in data['program']:
                            self._data = data
                        else:
                            logger.info(
                                'Could not get program id from %s',
                                data["program"],
                            )
                    else:
                        logger.info('Could not get program from %s', data)

    def __str__(self):
        return f'[{self.fn}: Version {self.version}]'

    @property
    def program(self):
        return self._data['program']

    @program.setter
    def program(self, program):
        self._data['program'] = program

    @property
    def version(self):
        return self._data.get('__version__', 0)

    @version.setter
    def version(self, version):
        self._data['__version__'] = version

    def write(self):
        with open(self.fn, 'w') as f:
            f.write(json.dumps(self._data, indent=4))

    def is_valid(self):
        return hasattr(self, '_data')

    @property
    def seasons(self):
        return {
            key: EntrySet(Entry.from_dict(entry) for entry in entries)
            for key, entries in self._data.items()
            if key not in NON_SEASON_FIELDS
        }

    @seasons.setter
    def seasons(self, seasons):
        keys = list(
            key for key in self._data.keys()
            if key not in NON_SEASON_FIELDS
        )
        for key in keys:
            del self._data[key]
        for key, entries in seasons.items():
            self._data[key] = [entry.to_dict() for entry in entries.sorted()]


class ProgramFetcher:
    pool = None

    def __init__(self, query=None, update=None, destination=None):
        if not destination:
            raise RuntimeError('Missing required destination parameter')
        self.query = query
        self.update = update
        self.destination = destination

    def get_programs(self):
        if self.query:
            return self.get_programs_by_query(self.query)
        return self.get_programs_to_update()

    def get_programs_by_query(self, queries):
        for query in queries:
            if query.isdigit():
                program_id = query
            else:
                program_id = self.get_program_id(query)
            r = requests.get(
                f'https://api.ruv.is/api/programs/program/{program_id}/all'
            )
            if r.ok:
                yield r.json()
            else:
                logger.warning(
                    f'Request for program {program_id} (query {query}) '
                    f'failed with status code {r.status_code}.'
                )

    def get_program_id(self, query):
        r = requests.get(
            f'https://api.ruv.is/api/programs/search/tv/{query}'
        )
        r.raise_for_status()
        programs = r.json()['programs']
        if not programs:
            raise RuntimeError(f'No programs found matching {query}')
        while True:
            for i, program in enumerate(programs):
                print(i + 1, ':', program['title'])
            selection = input('Select program: ')
            if selection.isdigit():
                selection = int(selection)
                if selection > 0 and selection <= len(programs):
                    return programs[selection - 1]['id']

    def get_all_program_infos(self):
        for fn in glob.glob(
            os.path.join(self.destination, '*', PROGRAM_INFO_FN)
        ):
            program_info = ProgramInfo(fn)
            if program_info.is_valid():
                yield program_info

    def get_programs_to_update(self):
        for program_info in self.get_all_program_infos():
            yield program_info.program



import os

DATE_FORMAT = '%Y/%m/%d'
DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
DATE_PART_LENGTH = 4 + 1 + 2 + 1 + 2
URL_TEMPLATE = (
    'http://smooth.ruv.cache.is/{openclose}/{date}/2400kbps/{fn}.mp4'
)


PROGRAM_INFO_FN = 'program_info.json'
NON_SEASON_FIELDS = ('program', '__version__')

CACHE_LOCATION = os.path.join(os.path.expanduser('~'), '.ruvdlcache')
# In case we change the cache setup, change the cache version value and we
# will invalidate all old cache.
CACHE_VERSION_KEY = '__cache_version__'
CACHE_VERSION = '1'

DEFAULT_VIDEO_DESTINATION = os.path.join(os.path.expanduser('~'), 'Videos/ruv')



import os

from setuptools import setup, find_packages

VERSION = '0.2.1.rc2'

ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(ROOT, 'README.md')) as f:
    readme = f.read()


setup(
    name='ruv-dl',
    version=VERSION,
    description='Downloader/organizer for ruv programs',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/sindrig/ruv-dl',
    author='Sindri Guðmundsson',
    author_email='sindrigudmundsson@gmail.com',
    license='MIT',
    packages=find_packages(),
    python_requires='>=3.6.0',
    install_requires=[
        'requests==2.22.0',
        'click==7.0',
    ],
    entry_points={
        'console_scripts': [
            'ruv-dl=ruv_dl:main',
        ],
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ]
)
